;*****************************************************************************
;*  Name    : umc_18FxxQ43.asm                                               *
;*  Author  : Jerry Messina                                                  *
;*          : based on original code by David John Barker                    *
;*  Date    : 24 FEB 2024                                                    *
;*  Version : UMC standalone bootloader support for 18FxxQ43 family          *
;*          : 1.0 - Initial release                                          *
;*  Notes   : Parts of this code based around the Tiny Bootloader            *
;*          : claudiu.chiculita@ugal.ro                                      *
;*          : http://www.etc.ugal.ro/cchiculita/software/picbootloader.htm   *  
;*****************************************************************************
; assemble using the SF copy of mpasmx along with the asm include files
; mpasmx include files can typically be found in 'swordfish\include\asm_includes' folder
; copy the appropriate include file for your device (ie 'p18F27Q43.inc') to the folder 
; containing this .asm file
;
; to assemble:
; - modify any settings in the USER CONFIGURATION SECTION below
; - run swordfish\bin\mpasmx.exe
;    using the Source File Name 'browse' button navigate to the folder containing this file 
;    select default settings for the other controls
;       Radix = Default
;       Warning Level = Default
;       Hex Output = Default
;       Generated Files = Error File, List File
;       Case Sensitive = unchecked
;       Macro Expansion = Default
;       Processor = Default
;       Extended Mode = unchecked
;    click 'Assemble'
;       this should generate the bootloader .hex file
;
; tested using:
;   mpasmx 5.94
;   mplabx 5.35
;   pickit4 toolpack 1.6.961
;   18F_Q DFP 1.14.237


   radix DEC          ; default radix is DECIMAL
   errorlevel -205    ; disable 'Found directive in column 1' warning
   list n=0           ; disable paging in list file
   
;=============================================================================
; USER CONFIGURATION SECTION
;=============================================================================
   processor                        18F27Q43
   #define DEVICE_CLOCK             64000000    ; for intosc select 8/16/32/64MHz
   #define INTOSC                   1           ; 0=ext osc, 1=internal HFINTOSC
   #define BAUDRATE                 115200      ; negotiated baudrate after startup
   
   ; it's recommended you don't change these
   #define DEFAULT_BAUDRATE         19200       ; initial startup baudrate (default is 19200)
   #define USE_BRGS                 1           ; high baudrate select (default is ON) 

  ; '#define TEST_UART' adds a poweron uart check routine that you can use
  ; to verify that the uart/clock is working properly. it sends an initial "U>"
  ; sequence and then loops echoing received chars. test with a terminal emulator
  ; or the SF IDE serial communicator (it uses the DEFAULT_BAUDRATE setting).
  ;#define TEST_UART 1        ; uncomment to add uart test code
  #ifdef TEST_UART
    messg "UART test code enabled"
  #endif

  ; the umc loader app only supports programming code space, so you can save some
  ; memory by excluding the EEPROM read/write functions. if you uncomment these
  ; you will have to increase LOADER_SIZE_MIN
  ; enable EEPROM READ
  ;#define ENABLE_EEPROM_READ 1
  ; enable EEPROM WRITE
  ;#define ENABLE_EEPROM_WRITE 1
  ; NOTE v1.0: eeprom routines are not tested...make sure EEPROM ops are not defined
  #ifdef ENABLE_EEPROM_READ
    messg "EEPROM operations not supported"
    #undefine ENABLE_EEPROM_READ
  #endif
  #ifdef ENABLE_EEPROM_WRITE
    messg "EEPROM operations not supported"
    #undefine ENABLE_EEPROM_WRITE
  #endif
  
;-----------------------------------------------------------------------------
; UART PORT/PIN selection
; modify this section to match your hardware configuration
; note that the Q43 PPS mapping only allows certain port selections:
; UART      2xQ43   4xQ43   5xQ43   
; UART1     B,C     B,C     C,F
; UART2     B,C     B,D     B,D
; UART3     A,B     A,B     A,F
; UART4 RX  B,C     B,D     B,D
; UART4 TX  A,B     A,D     A,D
; UART5 RX  A,C     A,C     A,F
; UART5 TX  B,C     B,C     C,F
; the following settings are for UART1 with PORTC RC6(TX out) and RC7(RX in) pins
;-----------------------------------------------------------------------------
#define UART_     1           ; UART select, 1-5
#define TX_PORT_  PORTC       ; TX output, PORTA-PORTF
#define TX_BIT_   6           ; TX output bit, 0-7
#define RX_PORT_  PORTC       ; RX input, PORTA-PORTF
#define RX_BIT_   7           ; RX input bit, 0-7

#if (UART_ == 0) || (UART_ > 5)
  error "UART_ selection error... 18FxxQ43 supports UART1-UART5"
#endif

;-----------------------------------------------------------------------------
; Q43 configuration fuse setup
;-----------------------------------------------------------------------------
#if (INTOSC == 0)            ; external xtal osc...
  config FEXTOSC = HS          ; HS (crystal oscillator) above 8 MHz; PFM set to high power
  config RSTOSC = EXTOSC       ; EXTOSC operating per FEXTOSC bits (default)
#endif
#if (INTOSC == 1)            ; internal osc...
  config FEXTOSC = OFF           ; Oscillator not enabled
  config RSTOSC = HFINTOSC_64MHZ ; HFINTOSC with HFFRQ = 64 MHz and CDIV = 1:1
#endif
  config CLKOUTEN = OFF           ; CLKOUT function is disabled
  config PR1WAY = OFF             ; PRLOCKED bit can be set and cleared repeatedly
  config CSWEN = ON               ; Writing to NOSC and NDIV is allowed
  config FCMEN = OFF              ; Fail-Safe Clock Monitor disabled
  config MCLRE = EXTMCLR          ; If LVP = 0, MCLR pin is MCLR; If LVP = 1, RE3 pin function is MCLR
  config PWRTS = PWRT_64          ; PWRT set at 64ms
  config MVECEN = OFF             ; Interrupt contoller does not use vector table to prioritze interrupts
  config IVT1WAY = OFF            ; IVTLOCKED bit can be cleared and set repeatedly
  config LPBOREN = OFF            ; Low-Power BOR disabled
  config BOREN = ON               ; Brown-out Reset enabled according to SBOREN
  config BORV = VBOR_1P9          ; Brown-out Reset Voltage (VBOR) set to 1.9V
  config ZCD = OFF                ; ZCD module is disabled. ZCD can be enabled by setting the ZCDSEN bit of ZCDCON
  config PPS1WAY = OFF            ; PPSLOCKED bit can be set and cleared repeatedly (subject to the unlock sequence)
  config STVREN = ON              ; Stack full/underflow will cause Reset
  config LVP = OFF                ; Low voltage programming disabled
  config XINST = OFF              ; Extended Instruction Set and Indexed Addressing Mode disabled
  config WDTCPS = WDTCPS_31       ; Divider ratio 1:65536; software control of WDTPS
  config WDTE = SWDTEN            ; WDT enabled/disabled by SWDTEN bit
  config WDTCWS = WDTCWS_7        ; window always open (100%); software control; keyed access not required
  config WDTCCS = SC              ; Software Control
  config BBSIZE = BBSIZE_512      ; Boot Block size is 512 words
  config BBEN = OFF               ; Boot block disabled
  config SAFEN = OFF              ; SAF disabled
  config DEBUG = OFF              ; Background Debugger disabled
  config WRTB = OFF               ; Boot Block not Write protected
  config WRTC = OFF               ; Configuration registers not Write protected
  config WRTD = OFF               ; Data EEPROM not Write protected
  config WRTSAF = OFF             ; SAF not Write Protected
  config WRTAPP = OFF             ; Application Block not write protected
  config CP = OFF                 ; PFM and Data EEPROM code protection disabled


; note: with umcbuild, this information is external, here it's included inline
; import the device information file - you may need to add devices to this file...
; #include devices.inc
;*****************************************************************************
;* device specific settings...                                               *
;* DEVICE_ID - the microchip device ID                                       *
;* DEVICE_SERIES - for example, 16F or 18F                                   *
;* DEVICE_ROM - size of device ROM (in bytes)                                *
;* DEVICE_BLOCK_ERASE - size of device erase block (in bytes)                *
;* DEVICE_BLOCK_WRITE - size of the device write block size (in bytes)       *
;* DEVICE_EEPROM - size of onboard EEPROM (in bytes)                         * 
;*****************************************************************************
#define DEVICE_IS_18F               0x04

; Q43 NOTE: the block erase and block write sizes are 256 bytes in these devices,
; but the UMC device info protocol only allows a byte value for the write size
; we'll write the data in two sets of 128 bytes
#ifdef  __18F25Q43
   messg "building for 18F25Q43"
   #include                         p18F25Q43.inc   
   #define DEVICE_ID                0xa2fe
   #define DEVICE_SERIES            DEVICE_IS_18F
   #define DEVICE_ROM               0x08000
   #define DEVICE_BLOCK_ERASE       256
   #define DEVICE_BLOCK_WRITE       128
   #define DEVICE_EEPROM            1024
   #define NVM_BUFFER_BANK          13
   #define NVM_BUFFER_SIZE          256
#endif
#ifdef  __18F26Q43
   messg "building for 18F26Q43"
   #include                         p18F26Q43.inc   
   #define DEVICE_ID                0xa2fd
   #define DEVICE_SERIES            DEVICE_IS_18F
   #define DEVICE_ROM               0x10000
   #define DEVICE_BLOCK_ERASE       256
   #define DEVICE_BLOCK_WRITE       128
   #define DEVICE_EEPROM            1024
   #define NVM_BUFFER_BANK          21
   #define NVM_BUFFER_SIZE          256
#endif
#ifdef  __18F27Q43
   messg "building for 18F27Q43"
   #include                         p18F27Q43.inc   
   #define DEVICE_ID                0xa2ea
   #define DEVICE_SERIES            DEVICE_IS_18F
   #define DEVICE_ROM               0x20000
   #define DEVICE_BLOCK_ERASE       256
   #define DEVICE_BLOCK_WRITE       128
   #define DEVICE_EEPROM            1024
   #define NVM_BUFFER_BANK          37
   #define NVM_BUFFER_SIZE          256
#endif
#ifdef  __18F45Q43
   messg "building for 18F45Q43"
   #include                         p18F45Q43.inc   
   #define DEVICE_ID                0xa2fc
   #define DEVICE_SERIES            DEVICE_IS_18F
   #define DEVICE_ROM               0x08000
   #define DEVICE_BLOCK_ERASE       256
   #define DEVICE_BLOCK_WRITE       128
   #define DEVICE_EEPROM            1024
   #define NVM_BUFFER_BANK          13
   #define NVM_BUFFER_SIZE          256
#endif
#ifdef  __18F46Q43
   messg "building for 18F46Q43"
   #include                         p18F46Q43.inc   
   #define DEVICE_ID                0xa2fb
   #define DEVICE_SERIES            DEVICE_IS_18F
   #define DEVICE_ROM               0x10000
   #define DEVICE_BLOCK_ERASE       256
   #define DEVICE_BLOCK_WRITE       128
   #define DEVICE_EEPROM            1024
   #define NVM_BUFFER_BANK          21
   #define NVM_BUFFER_SIZE          256
#endif
#ifdef  __18F47Q43
   messg "building for 18F47Q43"
   #include                         p18F47Q43.inc   
   #define DEVICE_ID                0xa2eb
   #define DEVICE_SERIES            DEVICE_IS_18F
   #define DEVICE_ROM               0x20000
   #define DEVICE_BLOCK_ERASE       256
   #define DEVICE_BLOCK_WRITE       128
   #define DEVICE_EEPROM            1024
   #define NVM_BUFFER_BANK          37
   #define NVM_BUFFER_SIZE          256
#endif
#ifdef  __18F55Q43
   messg "building for 18F55Q43"
   #include                         p18F55Q43.inc   
   #define DEVICE_ID                0xa2fa
   #define DEVICE_SERIES            DEVICE_IS_18F
   #define DEVICE_ROM               0x08000
   #define DEVICE_BLOCK_ERASE       256
   #define DEVICE_BLOCK_WRITE       128
   #define DEVICE_EEPROM            1024
   #define NVM_BUFFER_BANK          13
   #define NVM_BUFFER_SIZE          256
#endif
#ifdef  __18F56Q43
   messg "building for 18F56Q43"
   #include                         p18F56Q43.inc   
   #define DEVICE_ID                0xa2f9
   #define DEVICE_SERIES            DEVICE_IS_18F
   #define DEVICE_ROM               0x10000
   #define DEVICE_BLOCK_ERASE       256
   #define DEVICE_BLOCK_WRITE       128
   #define DEVICE_EEPROM            1024
   #define NVM_BUFFER_BANK          21
   #define NVM_BUFFER_SIZE          256
#endif
#ifdef  __18F57Q43
   messg "building for 18F57Q43"
   #include                         p18F57Q43.inc   
   #define DEVICE_ID                0xa2ec
   #define DEVICE_SERIES            DEVICE_IS_18F
   #define DEVICE_ROM               0x20000
   #define DEVICE_BLOCK_ERASE       256
   #define DEVICE_BLOCK_WRITE       128
   #define DEVICE_EEPROM            1024
   #define NVM_BUFFER_BANK          37
   #define NVM_BUFFER_SIZE          256
#endif

; 18FxxQ43 NVMCMD operations (NVMCON1)
; write/erase commands require NVM_UNLOCK
#define NVM_READ        0x00    ; NVM to NVMDATH/L (word)
#define NVM_READ_INCR   0x01    ; NVM to NVMDATH/L (word)
#define NVM_READ_PAGE   0x02    ; NVM to Buffer RAM (page)
#define NVM_WRITE       0x03    ; NVMDATH/L to NVM (word) 
#define NVM_WRITE_INCR  0x04    ; NVMDATH/L to NVM (word)
#define NVM_WRITE_PAGE  0x05    ; Buffer RAM to NVM (page)
#define NVM_ERASE_PAGE  0x06

; 18FxxQ43 access bank
#ifdef _accessbank        ; check for device file definition
  #define ACCESS_BANK   _accessbank
#else
  #define ACCESS_BANK   5
#endif

; define unused ports for 28-pin devices
#ifndef PORTD
  #define PORTD 0
#endif
#ifndef PORTF
  #define PORTF 0
#endif

; define registers based on user xx_PORT_ pin selection above
#if (TX_PORT_ == PORTA)
  #define TX_ANSELx_   ANSELA
  #define TX_SLRCONx_  SLRCONA
  #define TX_TRISx_    TRISA
#endif
#if (TX_PORT_ == PORTB)
  #define TX_ANSELx_   ANSELB
  #define TX_SLRCONx_  SLRCONB
  #define TX_TRISx_    TRISB
#endif
#if (TX_PORT_ == PORTC)
  #define TX_ANSELx_   ANSELC
  #define TX_SLRCONx_  SLRCONC
  #define TX_TRISx_    TRISC
#endif
#if (TX_PORT_ == PORTD)
  #define TX_ANSELx_   ANSELD
  #define TX_SLRCONx_  SLRCOND
  #define TX_TRISx_    TRISD
#endif
#if (TX_PORT_ == PORTE)
  error "PORTE unsupported for Q43 UART PPS"
  #define TX_ANSELx_   ANSELE
  #define TX_SLRCONx_  SLRCONE
  #define TX_TRISx_    TRISE
#endif
#if (TX_PORT_ == PORTF)
  #define TX_ANSELx_   ANSELF
  #define TX_SLRCONx_  SLRCONF
  #define TX_TRISx_    TRISF
#endif

#if (RX_PORT_ == PORTA)
  #define RX_ANSELx_   ANSELA
  #define RX_SLRCONx_  SLRCONA
  #define RX_TRISx_    TRISA
#endif
#if (RX_PORT_ == PORTB)
  #define RX_ANSELx_   ANSELB
  #define RX_SLRCONx_  SLRCONB
  #define RX_TRISx_    TRISB
#endif
#if (RX_PORT_ == PORTC)
  #define RX_ANSELx_   ANSELC
  #define RX_SLRCONx_  SLRCONC
  #define RX_TRISx_    TRISC
#endif
#if (RX_PORT_ == PORTD)
  #define RX_ANSELx_   ANSELD
  #define RX_SLRCONx_  SLRCOND
  #define RX_TRISx_    TRISD
#endif
#if (RX_PORT_ == PORTE)
  error "PORTE unsupported for Q43 UART PPS"
  #define RX_ANSELx_   ANSELE
  #define RX_SLRCONx_  SLRCONE
  #define RX_TRISx_    TRISE
#endif
#if (RX_PORT_ == PORTF)
  #define RX_ANSELx_   ANSELF
  #define RX_SLRCONx_  SLRCONF
  #define RX_TRISx_    TRISF
#endif

; alias Q43 UART registers and flags
#if (UART_ == 1)    ; UART1
  #define UxCON0      U1CON0        
  #define UxCON1      U1CON1
  #define UxCON2      U1CON2
  #define UxERRIR     U1ERRIR
  #define UxFIFO      U1FIFO
  #define UxBRGL      U1BRGL
  #define UxBRGH      U1BRGH
  #define UxRXB       U1RXB
  #define UxTXB       U1TXB
  #define PIR_        PIR4
  #define RCIF_       U1RXIF
  #define TXIF_       U1TXIF
  #define UxON_       U1ON
#endif  ; UART1
#if (UART_ == 2)    ; UART2
  #define UxCON0      U2CON0        
  #define UxCON1      U2CON1
  #define UxCON2      U2CON2
  #define UxERRIR     U2ERRIR
  #define UxFIFO      U2FIFO
  #define UxBRGL      U2BRGL
  #define UxBRGH      U2BRGH
  #define UxRXB       U2RXB
  #define UxTXB       U2TXB
  #define PIR_        PIR8
  #define RCIF_       U2RXIF
  #define TXIF_       U2TXIF
  #define UxON_       U2ON
#endif  ; UART2
#if (UART_ == 3)    ; UART3
  #define UxCON0      U3CON0        
  #define UxCON1      U3CON1
  #define UxCON2      U3CON2
  #define UxERRIR     U3ERRIR
  #define UxFIFO      U3FIFO
  #define UxBRGL      U3BRGL
  #define UxBRGH      U3BRGH
  #define UxRXB       U3RXB
  #define UxTXB       U3TXB
  #define PIR_        PIR9
  #define RCIF_       U3RXIF
  #define TXIF_       U3TXIF
  #define UxON_       U3ON
#endif  ; UART3
#if (UART_ == 4)    ; UART4
  #define UxCON0      U4CON0        
  #define UxCON1      U4CON1
  #define UxCON2      U4CON2
  #define UxERRIR     U4ERRIR
  #define UxFIFO      U4FIFO
  #define UxBRGL      U4BRGL
  #define UxBRGH      U4BRGH
  #define UxRXB       U4RXB
  #define UxTXB       U4TXB
  #define PIR_        PIR12
  #define RCIF_       U4RXIF
  #define TXIF_       U4TXIF
  #define UxON_       U4ON
#endif  ; UART4
#if (UART_ == 5)    ; UART5
  #define UxCON0      U5CON0        
  #define UxCON1      U5CON1
  #define UxCON2      U5CON2
  #define UxERRIR     U5ERRIR
  #define UxFIFO      U5FIFO
  #define UxBRGL      U5BRGL
  #define UxBRGH      U5BRGH
  #define UxRXB       U5RXB
  #define UxTXB       U5TXB
  #define PIR_        PIR13
  #define RCIF_       U5RXIF
  #define TXIF_       U5TXIF
  #define UxON_       U5ON
#endif  ; UART5

#define PPSLOCKED_   PPSLOCKED  ; PPSLOCK register bit
#define GIE_         GIE        ; INTCON0 register bit

;-----------------------------------------------------------------------------
; user init
;-----------------------------------------------------------------------------
UserInit macro
 #if (INTOSC == 1)
   ; setup Q43 internal osc
   banksel OSCCON1
   movlw  0x60               ; NOSC=110 (HFINTOSC), NDIV=1:1       
   movwf  OSCCON1
  #if (DEVICE_CLOCK == 1000000)
   movlw  0x00               ; 1MHz
  #endif
  #if (DEVICE_CLOCK == 2000000)
   movlw  0x01               ; 2MHz
  #endif
  #if (DEVICE_CLOCK == 4000000)
   movlw  0x02               ; 4MHz
  #endif
  #if (DEVICE_CLOCK == 8000000)
   movlw  0x03               ; 8MHz
  #endif
  #if (DEVICE_CLOCK == 12000000)
   movlw  0x04               ; 12MHz
  #endif
  #if (DEVICE_CLOCK == 16000000)
   movlw  0x05               ; 16MHz
  #endif
  #if (DEVICE_CLOCK == 32000000)
   movlw  0x06               ; 32MHz
  #endif
  #if (DEVICE_CLOCK == 48000000)
   movlw  0x07               ; 48MHz
  #endif
  #if (DEVICE_CLOCK == 64000000)
   movlw  0x08               ; 64MHz
  #endif
   banksel OSCFRQ
   movwf  OSCFRQ
 #endif   ; INTOSC

   ; set TX and RX pins to digital mode (ANSELx = 0)
   banksel  TX_ANSELx_
   bcf      TX_ANSELx_, TX_BIT_
   bcf      RX_ANSELx_, RX_BIT_
   
   ; set TX and RX pin normal slewrate (SLRCONx = 0)
   banksel  TX_SLRCONx_
   bcf      TX_SLRCONx_, TX_BIT_
   bcf      RX_SLRCONx_, RX_BIT_
   
   ; set uart TX and RX pin directions (these will over-ridden by the usart as req'd)
   banksel  TX_TRISx_
   bsf      TX_TRISx_, TX_BIT_
   banksel  RX_TRISx_
   bsf      RX_TRISx_, RX_BIT_

   ; PPS unlock
   ; disable interrupts
   bcf      INTCON0, GIE_
   banksel  PPSLOCK
   ; required sequence (next 4 instructions)
   movlw    0x55
   movwf    PPSLOCK
   movlw    0xAA
   movwf    PPSLOCK
   ; clear PPSLOCKED bit to enable writes (only a BCF instruction will work)
   bcf      PPSLOCK, PPSLOCKED_
   
   ; now assign uart pps pins
   banksel  RA0PPS              ; we assume all RxxPPS are in the same bank
  #if (UART_ == 1)
   movlw    ((RX_PORT_ - PORTA)<<3) | RX_BIT_           ; U1RXPPS usart 1 RX input
   movwf    U1RXPPS
   movlw    0x20                                        ; UART1 TX output
   movwf    RA0PPS + ((TX_PORT_ - PORTA)*8 + TX_BIT_)   ; RxxPPS
  #endif
  #if (UART_ == 2)
   movlw    ((RX_PORT_ - PORTA)<<3) | RX_BIT_           ; U2RXPPS usart 2 RX input
   movwf    U2RXPPS
   movlw    0x23                                        ; UART2 TX output
   movwf    RA0PPS + ((TX_PORT_ - PORTA)*8 + TX_BIT_)   ; RxxPPS
  #endif
  #if (UART_ == 3)
   movlw    ((RX_PORT_ - PORTA)<<3) | RX_BIT_           ; U3RXPPS usart 3 RX input
   movwf    U3RXPPS
   movlw    0x26                                        ; UART3 TX output
   movwf    RA0PPS + ((TX_PORT_ - PORTA)*8 + TX_BIT_)   ; RxxPPS
  #endif
  #if (UART_ == 4)
   movlw    ((RX_PORT_ - PORTA)<<3) | RX_BIT_           ; U4RXPPS usart 4 RX input
   movwf    U4RXPPS
   movlw    0x29                                        ; UART4 TX output
   movwf    RA0PPS + ((TX_PORT_ - PORTA)*8 + TX_BIT_)   ; RxxPPS
  #endif
  #if (UART_ == 5)
   movlw    ((RX_PORT_ - PORTA)<<3) | RX_BIT_           ; U5RXPPS usart 5 RX input
   movwf    U5RXPPS
   movlw    0x2C                                        ; UART5 TX output
   movwf    RA0PPS + ((TX_PORT_ - PORTA)*8 + TX_BIT_)   ; RxxPPS
  #endif
   
   ; set bank back to default access bank
   movlb    ACCESS_BANK
   endm

;=============================================================================
; END USER CONFIGURATION SECTION
;=============================================================================

;
; uart init
;
; baudrate multiplier is based on the setting of BRGS
;   BRGS=0 --> normal rate, /16
;   BRGS=1 --> high speed, /4
;   bauddiv = (FOSC / BAUDRATE / BRGS_DIV) - 1
#if (USE_BRGS == 0)
  #define MULT      16
#else
  #define MULT      4
#endif

; formula is (OSC / BAUDRATE / MULT) - 1. However, we need to round the result
; so that the SPBRG value is correct
#define MULT_DIV_2                  MULT / 2
#define DEFAULT_SPBRG_VALUE         ((DEVICE_CLOCK / DEFAULT_BAUDRATE + MULT_DIV_2) / MULT - 1)
#define USER_SPBRG_VALUE            ((DEVICE_CLOCK / BAUDRATE + MULT_DIV_2) / MULT - 1)

UartInit macro
   ; setup uart
   banksel  UxCON0
   movlw    0x00        ; serial port disabled
   movwf    UxCON1
   movlw    0x00        ; BRGS=0, TXEN=0, RXEN=0, Asynchronous 8-bit UART mode
   movwf    UxCON0
  #if (USE_BRGS == 1)
   bsf      UxCON0, BRGS   ; BRGS=1
  #endif
   movlw    0x00        ; normal polarity, 1 stop bit, flow control off
   movwf    UxCON2
   ; set default baudrate...
   movlw    (DEFAULT_SPBRG_VALUE) & 0xFF
   movwf    UxBRGL
   movlw    (DEFAULT_SPBRG_VALUE >> 8) & 0xFF
   movwf    UxBRGH

   ; init com to default - normally 19200 baud...
   movlw    0x22              ; clear rx and tx buffers
   movwf    UxFIFO
   bsf      UxCON0, RXEN      ; enable rx
   bsf      UxCON0, TXEN      ; enable tx
   bsf      UxCON1, UxON_     ; serial port enabled
   
   movlb    ACCESS_BANK
   endm

; set the minimum loader size...
; you may have to increase LOADER_SIZE_MIN for certain build options (test code, eeprom support, etc)
#define LOADER_SIZE_MIN             512
#define LOADER_SIZE                 LOADER_SIZE_MIN

; if the loader size is not an integer multiple of the erase block size, we
; need to set the loader size so it falls on an erase block boundary
#if (DEVICE_BLOCK_ERASE > 0)                ; if zero, then ignore
  #if ((LOADER_SIZE % DEVICE_BLOCK_ERASE) > 0)  ; if modulus not zero, we need a new loader size
    messg "LOADER_SIZE does not fall on an erase block boundary... it has been adjusted"
    #undefine LOADER_SIZE
    #define LOADER_SIZE  (DEVICE_BLOCK_ERASE * (LOADER_SIZE_MIN / DEVICE_BLOCK_ERASE + 1))
  #endif
#endif

; now create the loader and user code origin...
#define ORG_LOADER                  DEVICE_ROM - LOADER_SIZE
#define ORG_LOADER_USER             ORG_LOADER - 8

; loader commands - these are bit field identifiers
; data input values are shown in comments
#define CMD_EEPROM_READ             0x00              ; 0x01
#define CMD_EEPROM_WRITE            0x01              ; 0x02
#define CMD_ROM_ERASE               0x02              ; 0x04
#define CMD_ROM_READ                0x03              ; 0x08
#define CMD_ROM_WRITE               0x04              ; 0x10
#define CMD_BAUDRATE                0x05              ; 0x20

; addition information and commands...
#define FIRMWARE_VERSION            0x10
#define LOADER_CONNECT              0x80
#define LOADER_ACK                  LOADER_CONNECT
#define LOADER_GO                   0x81
#define LOADER_ERROR_CRC            0x8A
#define LOADER_ERROR_VERIFY         0x8B

 ; program variables
 #define BUFFER_SIZE  DEVICE_BLOCK_WRITE
 cblock (ACCESS_BANK * 256)
    CRC:        1
    Command:    1
    Ack:        1
    Bufcnt:     1
    EEVerify:   1
    Timeout_01: 1
    Timeout_02: 1
    Timeout_03: 1
    Buffer:     BUFFER_SIZE
 endc
 cblock (NVM_BUFFER_BANK*256)
    NVMBuffer:  NVM_BUFFER_SIZE  
 endc

   ; reset vector
   org    0x0000
   goto   LoaderStart

   ; relocated reset vector...
   org    ORG_LOADER_USER
LoaderStartUser:

   ; loader start
   org    ORG_LOADER
LoaderStart:
   ; user initialization macro...
   UserInit

   ; UART initialization macro...
   UartInit

  #ifdef TEST_UART
   rcall    TestUart
  #endif   

   ; can we connect to remote PC...
   rcall  ReadByte                ; read a byte
   sublw  LOADER_CONNECT          ; subtract connection ID
   bnz    LoaderStartUser         ; if not zero, run user code

   ; acknowledge...
   movlw  LOADER_CONNECT
   rcall  WriteByte

   ; wait for go before sending device information...
WaitForGo:
   rcall ReadByte
   sublw LOADER_GO
   bnz   WaitForGo

   ; firmware version...
   movlw  FIRMWARE_VERSION
   rcall  WriteByte

   ; baudrate...
   movlw  (BAUDRATE >> 16) & 0xFF
   rcall  WriteByte
   movlw  (BAUDRATE >> 8) & 0xFF
   rcall  WriteByte
   movlw  (BAUDRATE) & 0xFF
   rcall  WriteByte

   ; device series...
   movlw  (DEVICE_SERIES) & 0xFF
   rcall  WriteByte

   ; device ID...
   movlw  (DEVICE_ID >> 8) & 0xFF
   rcall  WriteByte
   movlw  (DEVICE_ID) & 0xFF
   rcall  WriteByte

   ; the loader origin - this defines the loader protection point
   ; note that the PC application will offset this value by 8 bytes 
   ; to take into account the relocatable user code address
   movlw  (ORG_LOADER >> 16) & 0xFF
   rcall  WriteByte
   movlw  (ORG_LOADER >> 8) & 0xFF
   rcall  WriteByte
   movlw  (ORG_LOADER) & 0xFF
   rcall  WriteByte

   ; erase block size...
   movlw  (DEVICE_BLOCK_ERASE >> 8) & 0xFF
   rcall  WriteByte
   movlw  (DEVICE_BLOCK_ERASE) & 0xFF
   rcall  WriteByte

   ; write block size...
   movlw  (DEVICE_BLOCK_WRITE) & 0xFF
   rcall  WriteByte

   ; init ack
   movlw  LOADER_ACK
   movwf  Ack

; main program loop
LoaderMain:

   ; get command...
   rcall  ReadByte                ; read a byte
   movwf  Command                 ; set command
   clrf   CRC                     ; clear CRC

   ; switch to user defined baudrate - the loader starts in the default
   ; 19200 baud - this enables you to crank up the speed after initial
   ; startup handshaking has taken place...
  #if (BAUDRATE != DEFAULT_BAUDRATE)
   btfss  Command, CMD_BAUDRATE   ; check command
   bra    NoBaudrate              ; no baudrate change

   ; set baudrate...
   banksel   UxBRGL
   movlw     (USER_SPBRG_VALUE) & 0xFF
   movwf     UxBRGL
   movlw     (USER_SPBRG_VALUE >> 8) & 0xFF
   movwf     UxBRGH
   movlb    ACCESS_BANK
   bra    LoaderMain              ; no ack (settle time) loop back
  #endif
NoBaudrate:

   ; read upper byte of ROM address and high byte of EEPROM address...
   rcall  ReadByte                ; read a byte
   movwf  TBLPTRU                 ; move into table pointer upper
  #if (DEVICE_EEPROM > 256)       ; if we have extended EEPROM
   banksel NVMCON1
   movwf  NVMADRH                 ; then move value into EEPROM address high
   movlb    ACCESS_BANK
  #endif

   ; read high byte of ROM address and low byte of EEPROM addresss...
   rcall  ReadByte                ; read a byte
   movwf  TBLPTRH                 ; move into table pointer high
  #if (DEVICE_EEPROM > 0)         ; if we have onboard EEPROM
   banksel NVMCON1
   movwf  NVMADRL                 ; then move value into EEPROM address low
   movlb    ACCESS_BANK
  #endif

   ; read low byte of ROM address and data EEPROM byte...
   rcall  ReadByte                ; read a byte
   movwf  TBLPTRL                 ; move into table pointer low
  #if (DEVICE_EEPROM > 0)         ; if we have onboard EEPROM
   banksel NVMCON1
   movwf  NVMDAT                  ; then move value into EEPROM data
   movlb    ACCESS_BANK
   movwf  EEVerify                ; store a copy - used to verify the written data
  #endif

   ; EEPROM READ...
  #ifdef ENABLE_EEPROM_READ
   #if (DEVICE_EEPROM > 0)         ; if device has onboard EEPROM
   btfsc  Command, CMD_EEPROM_READ ; check command
   rcall  EEPROMRead               ; execute EEPROM read
   #endif
  #endif

   ; EEPROM WRITE...
  #ifdef ENABLE_EEPROM_WRITE
   #if (DEVICE_EEPROM > 0)         ; if device has onboard EEPROM
   btfsc  Command,CMD_EEPROM_WRITE ; check command
   rcall  EEPROMWrite              ; execute EEPROM write
   #endif
  #endif

   ; CODE ERASE...
   btfsc  Command, CMD_ROM_ERASE  ; check command
   rcall  ROMErase                ; execute program code erase

   ; CODE READ...
   btfsc  Command, CMD_ROM_READ   ; check command
   rcall  ROMRead                 ; execute program code read

   ; CODE WRITE...
   btfsc  Command, CMD_ROM_WRITE  ; check command
   rcall  ROMWrite                ; execute program code write

   ; write ack...
   movf   Ack, W                  ; ack byte to w reg
   rcall  WriteByte               ; send
   bra    LoaderMain              ; then loop back

; verify error...
VerifyError:
   movlb    ACCESS_BANK
   movlw  LOADER_ERROR_VERIFY     ; set flag
   rcall  WriteByte               ; send ack
   bra    LoaderMain              ; loop back to main

; CRC error...
CRCError:
   movlb    ACCESS_BANK
   movlw  LOADER_ERROR_CRC        ; set flag
   rcall  WriteByte               ; send ack
   bra    LoaderMain              ; loop back to main

; execute user code...
RunUserCode:
   clrf     TBLPTRU                 ; clear table pointer upper
   clrf     STKPTR                  ; reset stack pointer
   banksel  UxCON0
   clrf     UxCON0                ; reset uart regs
   clrf     UxCON1
   clrf     UxCON2
   movlb    ACCESS_BANK
   goto   LoaderStartUser         ; execute user code

;*****************************************************************************
; Name    : WriteByte                                                        *
; Purpose : Write a byte in WREG to the hardware USART                       *
;*****************************************************************************
WriteByte:
   clrwdt                         ; clear WDT
   banksel  UxERRIR
   btfss    UxERRIR, TXMTIF       ; can we send data (10-08-2010)
   bra      WriteByte             ; no - loop back
   movwf    UxTXB                 ; send data
   movlb    ACCESS_BANK
   return

;*****************************************************************************
; Name    : ReadByte                                                         *
; Purpose : Read a byte from the hardware USART. Return in WREG              *
;         : If a timeout occurs, the routine will branch to 'RunUserCode'    *
;         : TIMEOUT                                                          *
;         : Assume a clock of 20MHz, then one instruction takes              *
;         : 1 / 20 000 000 * 4 = 2us to execute                              *
;         : Timeout_03 loop costs 6 cycles                                   *
;         : Timeout_02 loop costs 4 cycles                                   *
;         : Timeout_01 loop costs 4 cycles                                   *
;         : 256 * 6          = 1536 cycles                                   *
;         : 256 * 1536 + 4   = 393220 cycles                                 *
;         : 20 * 393220 + 4  = 7864404 cycles                                *
;         : 7864404 * 2us = 1.57 seconds - this is constant through varying  *
;         : clock speeds                                                     *
;*****************************************************************************
ReadByte:
   movlb    ACCESS_BANK
   movlw    (DEVICE_CLOCK / 1000000)
   movwf    Timeout_01
TimeoutLoop_01:                    ;  4 cycle cost, excluding inner...
   clrf     Timeout_02             ; (1)
TimeoutLoop_02:                    ;  4 cycle cost, excluding inner...
   clrf     Timeout_03             ; (1)
TimeoutLoop_03:                    ;  6 cycle cost - if nothing received...
   clrwdt                          ; (1)
   banksel  PIR_
   btfss    PIR_, RCIF_             ; (1) if not skipped
   bra      NotReceived             ; (2)
   banksel  UxRXB
   movf     UxRXB, W                ; return in WREG
   addwf    CRC,F                   ; calculate the CRC checksum
   movlb    ACCESS_BANK
   return
NotReceived:
   movlb    ACCESS_BANK
   decfsz   Timeout_03,F            ; (1) if not skipped
   bra      TimeoutLoop_03          ; (2)
   decfsz   Timeout_02,F            ; (1) if not skipped
   bra      TimeoutLoop_02          ; (2)
   decfsz   Timeout_01,F            ; (1) if not skipped
   bra      TimeoutLoop_01          ; (2)
   bra      RunUserCode


;*****************************************************************************
; Name    : TestUart                                                         *
; Purpose : uart loopback test                                               *
;*****************************************************************************
; uart test routine (conditional)
#ifdef TEST_UART
TestUart:
   movlw  0x55           ; display 'U>'
   rcall  WriteByte
   movlw  0x3e
   rcall  WriteByte
echo_loop:                ; read and echo all chars
   rcall  GetByte
   rcall  WriteByte
   goto   echo_loop
   return
 
GetByte:
wait_char:
   clrwdt
   banksel  UxERRIR
   btfsc    UxERRIR, RXFOIF       ; check for overrun error
   bra      handle_oerr
   banksel  PIR_
   btfss    PIR_, RCIF_             ; (1) if not skipped
   bra      wait_char
   banksel  UxRXB
   movf     UxRXB, W               ; return in WREG
   movlb    ACCESS_BANK
   return
handle_oerr:
   ; the receive overflow condition is cleared by reading the UxRXB register
   ; and clearing the RXFOIF bit
   movf     UxRXB, W
   bcf      UxERRIR, RXFOIF
   bra      wait_char
#endif       ; ifdef TEST_UART

;*****************************************************************************
; Name    : EEPROMRead
; Purpose : Read a single byte from onboard EEPROM
;   assumes NVMADRU/NVMADRH/NVMADRL has been set
;*****************************************************************************
#ifdef ENABLE_EEPROM_READ
#if (DEVICE_EEPROM > 0)
EEPROMRead:
   banksel NVMCON1
   movlw  NVM_READ                ; set the byte read command
   movwf  NVMCON1
   bsf    NVMCON0, NVMGO          ; start byte read
EEPROMRead_Wait:
   btfsc  NVMCON0, NVMGO          ; wait for the read operation to complete
   bra    EEPROMRead_Wait
   movf   NVMDATL, W              ; store the read value in W 
   movlb    ACCESS_BANK
   rcall  WriteByte               ; send byte read (WREG) 
   return
#endif
#endif

;*****************************************************************************
; Name    : EEPROMWrite
; Purpose : Write a single byte of data to onboard EEPROM
;  assumes NVMADRU/NVMADRH/NVMADRL has been set, data is already in NVMDATL
;*****************************************************************************
#ifdef ENABLE_EEPROM_WRITE
#if (DEVICE_EEPROM > 0)
EEPROMWrite:
   ; CRC check...
   rcall  ReadByte                ; read a byte
   tstfsz CRC                     ; test for zero
   bra    CRCError                ; not zero, so error

   banksel NVMCON1
   movlw  NVM_WRITE               ; set the byte write command
   movwf  NVMCON1
   rcall  NVMUnlock               ; required write sequence

   ; now verify the data
   banksel NVMCON1
   movlw  NVM_READ                ; set the byte read command
   movwf  NVMCON1
   bsf    NVMCON0, NVMGO          ; start byte read
EEPROMVerify_Wait:
   btfsc  NVMCON0, NVMGO          ; wait for the read operation to complete
   bra    EEPROMVerify_Wait
   movf   EEVerify, W             ; data byte received
   cpfseq NVMDATL                 ; compare
   bra    VerifyError             ; failed - set ack
   movlb    ACCESS_BANK
   return
#endif
#endif

;*****************************************************************************
; Name    : ROMErase                                                         *
; Purpose : Erase a block of device ROM                                      *
;*****************************************************************************
ROMErase:
   ; set NVMADR from TBLPTR regs
   movff  TBLPTRU, NVMADRU
   movff  TBLPTRH, NVMADRH
   movff  TBLPTRL, NVMADRL       
   
   banksel NVMCON1
   movlw  NVM_ERASE_PAGE          ; set the page erase command
   movwf  NVMCON1
   rcall  NVMUnlock               ; required write sequence
   movlb    ACCESS_BANK
   return

;*****************************************************************************
; Name    : ROMRead                                                          *
; Purpose : Read a single byte from ROM                                      *
;*****************************************************************************
ROMRead:
   movlb    ACCESS_BANK
   tblrd*+                        ; perform table read (low byte)
   movf   TABLAT, W               ; WREG <- TABLAT
   rcall  WriteByte               ; send the byte
   return

;*****************************************************************************
; Name    : ROMWrite                                                         *
; Purpose : Write data held in data buffer to ROM                            *
;*****************************************************************************
ROMWrite:
   movlb    ACCESS_BANK
   lfsr   0, Buffer               ; load FSR0 with buffer address
   movlw  BUFFER_SIZE             ; load buffer size
   movwf  Bufcnt                   ; into Bufcnt

ReadData:
   rcall  ReadByte                ; read a byte from the UART
   movwf  POSTINC0                ; save to buffer
   decfsz Bufcnt                   ; dec Bufcnt
   bra    ReadData                ; loop back if not zero

   ; CRC check...
   rcall  ReadByte                ; read a byte
   tstfsz CRC                     ; test for zero
   bra    CRCError                ; not zero, so error

   ; set NVMADR page address from TBLPTR regs
   banksel NVMADRL
   movff  TBLPTRU, NVMADRU
   movff  TBLPTRH, NVMADRH
   movff  TBLPTRL, NVMADRL
;   clrf   NVMADRL
   
   ; read current NVM data into the NVM buffer ram
   banksel NVMCON1
   movlw  NVM_READ_PAGE          ; set the page read command
   movwf  NVMCON1
   bsf    NVMCON0, NVMGO        ; start the operation
ReadNVMpage_Wait:
   btfsc  NVMCON0, NVMGO        ; wait for the operation to complete
   bra    ReadNVMpage_Wait
   clrf   NVMCON1               ; disable writes to memory (NVMCMD=NVM_READ)

   ; copy the buffer data into the NVM buffer
   lfsr   0, Buffer               ; load buffer address
   lfsr   1, NVMBuffer
   movff  NVMADRL, FSR1L          ; adjust for first 128 bytes or second 128 bytes of page
   movlw  BUFFER_SIZE             ; loader buffer size
   movwf  Bufcnt                   ; into Bufcnt

LoadData:
   movff  POSTINC0, POSTINC1      ; copy data from buffer to NVMbuffer
   decfsz Bufcnt                  ; dec Bufcnt
   bra    LoadData                ; loop back if not zero

   ; write the data...
   movlw  NVM_WRITE_PAGE          ; set the page write command
   movwf  NVMCON1
   rcall  NVMUnlock               ; required write sequence

   ; verify the write...
   lfsr   0, Buffer               ; load buffer address
   movlw  BUFFER_SIZE             ; load buffer size
   movwf  Bufcnt                   ; into Bufcnt

VerifyData:
   tblrd*+                        ; read data from ROM
   movf   POSTINC0, W             ; read data from RAM
   cpfseq TABLAT                  ; compare values
   bra    VerifyError             ; failed, set ack
   decfsz Bufcnt                   ; dec Bufcnt
   bra    VerifyData              ; loop back if not zero

   movlb    ACCESS_BANK
   return

;*****************************************************************************
; Name    : NVMUnlock                                                        *
; Purpose : A required write sequence for EEPROM and CODE writes             *
;*****************************************************************************
NVMUnlock:
   banksel NVMLOCK              ; send UNLOCK sequence
   movlw  0x55
   movwf  NVMLOCK
   movlw  0xAA
   movwf  NVMLOCK

   bsf    NVMCON0, NVMGO        ; start the operation

NVMUnlock_Wait:
   btfsc  NVMCON0, NVMGO        ; wait for the operation to complete
   bra    NVMUnlock_Wait

   clrf   NVMCON1               ; disable writes to memory (NVMCMD=NVM_READ)
   return

; check to make sure program fits in the alloted space
#if ($ > DEVICE_ROM)
  error "program code exceeds flash size. change LOADER_SIZE_MIN"
#endif
   end
